--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local AddonName, NS = ...

local L, OptionsFrame

local floor = math.floor
local clip = string.trim
local band = bit.band

NS.UIW, NS.UIH = floor(UIParent:GetWidth()), floor(UIParent:GetHeight())

local Path = "Interface\\AddOns\\"..AddonName.."\\"
NS.Font = Path.."Fonts\\Charcoal.ttf"
NS.FontSize = 14
NS.TexturePath = Path.."Textures\\"

NS.Lib = NS.FizzleLib or FizzleLib:GetLib(AddonName)

NS.DebugPrint = FIZZLEDEBUG and FIZZLEDEBUG(AddonName) or function() end

NS.Versions = {
	AutoBar = {},
	Dominos = {},
	DiscordUnitFrames = {},
	Quartz = {},
	SmartBuff = {},
	TipTac = {},
	SmartBuff = {},
}

local function UpdateVersions()
	if NS:IsValidVersion(NS.WOW_Project.Cataclysm_Classic) then
		NS.Versions.Dominos[WOW_PROJECT_ID] = { scale = 0.78 }
		NS.Versions.Quartz[WOW_PROJECT_ID] = { scale = 0.71 }
		NS.Versions.SmartBuff[WOW_PROJECT_ID] = { scale = 0.71 }
		NS.Versions.TipTac[WOW_PROJECT_ID] = { scale = 0.8 }
		NS.Versions.SmartBuff[WOW_PROJECT_ID] = { scale = 0.74 }

 		NS.Versions.AutoBar[WOW_PROJECT_ID] = { scale = 0.65, mainscale = 0.53, padding=5, mainyos = -1 }
		NS.Versions.DiscordUnitFrames[WOW_PROJECT_ID] = { scale = 0.715 }
	end
	if NS:IsValidVersion(NS.WOW_Project.WrathClassic) then
		NS.Versions.Dominos[WOW_PROJECT_ID] = { scale = 0.78 }
		NS.Versions.Quartz[WOW_PROJECT_ID] = { scale = 0.71 }
		NS.Versions.SmartBuff[WOW_PROJECT_ID] = { scale = 0.71 }
		NS.Versions.TipTac[WOW_PROJECT_ID] = { scale = 0.8 }
		NS.Versions.SmartBuff[WOW_PROJECT_ID] = { scale = 0.74 }

 		NS.Versions.AutoBar[WOW_PROJECT_ID] = { scale = 0.65, mainscale = 0.53, padding=5, mainyos = -1 }
		NS.Versions.DiscordUnitFrames[WOW_PROJECT_ID] = { scale = 0.78 }
	end
	if NS:IsValidVersion(NS.WOW_Project.Classic, NS.WOW_Project.Burning_Crusade_Classic) then
		NS.Versions[WOW_PROJECT_ID] = { offset = 0.73, }
		NS.Versions.Dominos[WOW_PROJECT_ID] = { scale = 0.71 }
		NS.Versions.Quartz[WOW_PROJECT_ID] = { scale = 0.71 }
		NS.Versions.SmartBuff[WOW_PROJECT_ID] = { scale = 0.71 }
		NS.Versions.TipTac[WOW_PROJECT_ID] = { scale = 0.8 }
		NS.Versions.SmartBuff[WOW_PROJECT_ID] = { scale = 0.74 }
	end
	if NS:IsValidVersion(NS.WOW_Project.Classic) then
 		NS.Versions.AutoBar[WOW_PROJECT_ID] = { scale = 0.71, mainscale = 0.61, padding=5, mainyos = -3 }
		NS.Versions.DiscordUnitFrames[WOW_PROJECT_ID] = { scale = 0.71 }
	end
	if NS:IsValidVersion(NS.WOW_Project.Burning_Crusade_Classic) then
 		NS.Versions.AutoBar[WOW_PROJECT_ID] = { scale = 0.65, mainscale = 0.60, padding=5, mainyos = -5 }
		NS.Versions.DiscordUnitFrames[WOW_PROJECT_ID] = { scale = 0.715 }
	end
	if NS:IsValidVersion(WOW_PROJECT_MAINLINE) then
		if GetExpansionLevel() >= 8 then
			NS.Versions.AutoBar[WOW_PROJECT_ID] = { scale = 0.74, mainscale = 0.62, padding=5, mainyos = -1 }
			NS.Versions.Dominos[WOW_PROJECT_ID] = { scale = 0.84 }
		else
			NS.Versions.AutoBar[WOW_PROJECT_ID] = { scale = 0.88, mainscale = 0.78, padding=5, mainyos = -1 }
		end
	end
end


--To defaults maybe?? NS.LayoutColor = { r=0.22, g=0.22, b=0.22 }
--To defaults maybe?? NS.LayoutAlpha = 0.3

function NS:GetAceProfile()
	return UnitName("player") .. " - " .. GetRealmName()
end

NS.AceDb = LibStub:GetLibrary("AceDB-3.0", true)
NS.AceDBOptions = LibStub:GetLibrary("AceDBOptions-3.0", true)
NS.AceAddons = LibStub:GetLibrary("AceAddon-3.0", true)

local ConfigFuncs = { addons={} }

-- addonlistexclude=exclude from list of Addons. reloadonchange = check and highlight for reload if inactive status has changed
function NS:AddConfig(order, addonname, displayname, configfunc, assignfunc, addonlistexclude, reloadonchange)
	if not ConfigFuncs[order] then
		ConfigFuncs[order] = {}
	end
	tinsert(ConfigFuncs[order], { config=configfunc, assign=assignfunc, addon=addonname})
	ConfigFuncs.addons[addonname] = { display=displayname, order=order, index=#ConfigFuncs[order], exclude=addonlistexclude, reloadonchange=reloadonchange}
end

function NS:GetConfigFuncs()
	return ConfigFuncs
end

------------------------------------------------------------------------------------------------------------------------------------
--							Settings Updates							  --
------------------------------------------------------------------------------------------------------------------------------------

local AddonsInitialised, AddonsAssigned

-- Add Latest #NS.NewSettings to main addon when changed
local function UpdateNewSettings(lastnewsetting, reset)
	local NewSettingsControlTable = NS:GetOption("NewSettings")
	local LastUpdateForGlobal = NS:GetSettings().Version or 0
	local LastUpdateForCharacter = NS:GetSettings("Profile").Version or 0
	local LastGlobalUpdate, LastCharacterUpdate = 0, 0
	for i=lastnewsetting, 1, -1 do
		if not LastGlobalUpdate and NewSettingsControlTable[i] == false then
			LastGlobalUpdate = i
		end			
		if not LastCharacterUpdate and NewSettingsControlTable[i] == true then
			LastCharacterUpdate = i
		end
		if LastGlobalUpdate > 0 and LastCharacterUpdate > 0 then 
			break 
		end
	end
	if LastUpdateForGlobal > 0 or LastUpdateForCharacter > 0 then
		if LastUpdateForCharacter == LastCharacterUpdate and LastUpdateForGlobal == LastGlobalUpdate then
			return
		end
	end
	local Start
	if LastUpdateForCharacter < LastGlobalUpdate then
		Start = LastUpdateForCharacter
	else
		Start = LastUpdateForGlobal
	end
	Start = Start + 1
	local SaveInit, SaveOpen, CharacterVersion
	for i=Start, lastnewsetting do
		local UpdateIsGlobal = not NewSettingsControlTable[i]
		if (UpdateIsGlobal and i > LastUpdateForGlobal) or (not UpdateIsGlobal and i > LastUpdateForCharacter) then
			local Init, Open = NS.NewSettings[i](LastGlobalUpdate, lastnewsetting, AddonsInitialised, AddonsAssigned, reset)
			SaveInit = SaveInit or Init
			SaveOpen = SaveOpen or Open
			if not UpdateIsGlobal then
				NS:GetSettings("Profile").Version = i
			else
				NS:GetSettings().Version = i
				NS:GetSettings("Profile").Version = NS:GetSettings("Profile").Version or 0
				for i = #NewSettingsControlTable, 1, -1 do
					if NewSettingsControlTable[i] == false then
						NS:GetSettings("Profile").Version = i
						break
					end
				end
			end
		end
	end
	return SaveInit, SaveOpen
end

function NS:InitialiseAddons()
	for i=1, #ConfigFuncs do
		for f=1, #ConfigFuncs[i] do
			if ConfigFuncs[i][f].config then
				ConfigFuncs[i][f].config()
			end
		end
	end
	AddonsInitialised = true
--	local settings = NS:GetSettings("Meters") -- Maybe here for initialising Meters if globsl config.
end

local function AssignAddons()
	local Settings = NS:GetSettings("Addons")
	for i=1, #ConfigFuncs do
		for f=1, #ConfigFuncs[i] do
			if not Settings[ConfigFuncs[i][f].addon] then
				Settings[ConfigFuncs[i][f].addon] = {}
			end
			if ConfigFuncs[i][f].assign then
				ConfigFuncs[i][f].assign()
			end
		end
	end
	AddonsAssigned = true
end

local function InitiaseData()
	local Metrics = NS:GetOption("Metrics")
	local x=Metrics.UIW/4
	local y = -Metrics.UIH/5
	if TalkingHeadFrame then
		NS.TalkingHeadPos = {
			{ text=L.Top, value="TOP", x=0, y=y*0.7 }, 
			{ text=L.TopLeft, value="TOPLEFT", x=x, y=y }, 
			{ text=L.TopRight, value="TOPRIGHT", x=-x, y=y }, 
			{ text=L.BottomRight, value="BOTTOMRIGHT", x=0, y=0 }, 
			{ text=L.Ignore, value="IGNORE", x=-0, y=0 }, 
		}
	end
end

------------------------------------------------------------------------------------------------------------------------------------
-- 						System Initialisation								  --
------------------------------------------------------------------------------------------------------------------------------------
function FauxMazzleConfig(getfuncs, setuptype, transition)
	if not NS.GetOption then
		if not getfuncs then
			print(AddonName..":", "No function info. passed!")
			return
		end
		NS.GetOption = getfuncs
		NS.GetSettings = NS:GetOption("GetSettings")
		NS.GetCharacterSettings = NS:GetOption("GetCharacterSettings")
		NS.GetFrame = NS:GetOption("GetFrame")
		NS.IsValidVersion = NS:GetOption("IsValidVersion")
		NS.WOW_Project = NS:GetOption("WowProjects")
		NS.TEXTS = NS:GetOption("Texts")
		L = NS.TEXTS
		NS:SetLocale(L)
		if not FAUXMAZZLEConfig then
			FAUXMAZZLEConfig = {}
		end
		NS.ConfigOptions = FAUXMAZZLEConfig
		UpdateVersions()
		InitiaseData()
	end
	local UpdateAddons, Reset
--	if band(setuptype, NS:GetOption("SetupType").Reset) > 0 then
	if band(setuptype, NS:GetOption("SetupType").Reset) == NS:GetOption("SetupType").Reset then
		Reset = true
	end
--	if band(setuptype, NS:GetOption('SetupType').Initialise) > 0 then
	if band(setuptype, NS:GetOption('SetupType').Initialise) == NS:GetOption('SetupType').Initialise then
		NS:DefaultsInitialiseGlobal()
	end
--	if band(setuptype, NS:GetOption('SetupType').InitialiseAddons) > 0 then
	if band(setuptype, NS:GetOption('SetupType').InitialiseAddons) == NS:GetOption('SetupType').InitialiseAddons then
		NS:InitialiseAddons()
		NS:GetSettings("Profile").Version = NS:GetSettings("Profile").Version or 0
		local NewSettingsControlTable = NS:GetOption("NewSettings")
		for i = #NewSettingsControlTable, 1, -1 do
			if NewSettingsControlTable[i] == false then
				NS:GetSettings("Profile").Version = i
				break
			end
		end
		return
	end
	local Settings = NS:GetSettings()
	if not NS.Profile then
		NS.Profile = NS:GetOption("Profile")
	end
--	if band(setuptype, NS:GetOption("SetupType").Character) > 0 then
	if band(setuptype, NS:GetOption("SetupType").Character) == NS:GetOption("SetupType").Character then
		NS:DefaultsInitialiseCharacter(NS.Profile)
		if not Reset then
			DiscordLib_RegisterInitFunc(AssignAddons)
		else
			AssignAddons()
			setuptype = setuptype + NS:GetOption("SetupType").NewSettings
			FauxMazzleFrame:GetScript("OnEvent")(FauxMazzleFrame, "CHANNEL_UI_UPDATE")
		end
		local NewSettingsControlTable = NS:GetOption("NewSettings")
		for i = #NewSettingsControlTable, 1, -1 do
			if NewSettingsControlTable[i] == true then
				NS:GetSettings().Version = i
				break
			end
		end
	end
	Settings = NS:GetSettings("Profile")
--	if band(setuptype, NS:GetOption("SetupType").NewSettings) > 0 then
	if band(setuptype, NS:GetOption("SetupType").NewSettings) == NS:GetOption("SetupType").NewSettings then
		local NewSettingsControlTable = NS:GetOption("NewSettings")
		local LastNewSetting = #NewSettingsControlTable
		if FIZZLEDEBUG and #NS.NewSettings ~= #NewSettingsControlTable then
			NS.DebugPrint("Missmatched New Settings Tables!", "Ignore if testing!")
		end
		UpdateAddons, NS:GetCharacterSettings().OpenConfig = UpdateNewSettings(LastNewSetting, Reset) -- In case we're testing and have enable "latest" New Setting 
	end
	if transition then
		NS:TransitionxPac()
	end
	if UpdateAddons then
--		Reset all the addons and
--		NS:DefaultsAddonConfigInitialiseAll()
	end
--	if band(setuptype, NS:GetOption('SetupType').OpenDialog) > 0 then
	if band(setuptype, NS:GetOption('SetupType').OpenDialog) == NS:GetOption('SetupType').OpenDialog then
--		local InitVal = band(setuptype, NS:GetOption('SetupType').HighlightReload)
		local InitVal = band(setuptype, NS:GetOption('SetupType').HighlightReload)
		if InitVal == NS:GetOption('SetupType').HighlightReload then
			InitVal = true
		else
			InitVal = false
		end
--		if not OptionsFrame then -- New options session, create the frame
			OptionsFrame = NS:LoadUI(InitVal)
--		end
--		OptionsFrame.HighlightReload = InitVal
--		OptionsFrame:SetShown(not OptionsFrame:IsShown())
	end
end

function NS:AdjustSettings(tbl, func, ...)
	for key, value in pairs(tbl) do
		func(tbl, key, value, ...)
		if type(value) == "table" then
			NS:AdjustSettings(value, func, ...)
		end
	end
end

---------------------------------- Chat Functions ----------------------------------
function NS:GetChatWindowName(channelid, defaultname)
	local name = GetChatWindowInfo(channelid)
	if not name or clip(name) == "" then
		name = defaultname and defaultname or L.Unused
	end
	return name
end

function NS:SetChatWindowName(id, name)
	local Frame = _G["ChatFrame"..id]
	SetChatWindowName(id, name)
	Frame.name = name
	if Frame.minFrame then
		Frame.minFrame:SetText(name)
	end
end

--function NS:ChannelUpdateDelayed(chatframeid, action, channelnum, channelname, ...)
function NS:ChannelUpdateDelayed(chatframeid, action, channelname, ...)-- Chaned to use channel names rather than numbers
	local Settings = NS:GetSettings("Chat")
	local IsDisabledChannel = NS:GetOption("IsDisabledChannel")    -- (it's possible they might changedepending on order added to a ChatFrame???)
--	local Disabled = IsDisabledChannel(channelnum, ...)
	local Disabled = IsDisabledChannel(channelname, ...)
	if Disabled then
		if not Settings.delaychannelupdate then
			Settings.delaychannelupdate = {}
		end
		if not Settings.delaychannelupdate[action] then
			Settings.delaychannelupdate[action] = {}
		end
--		tinsert(Settings.delaychannelupdate[action], { id=chatframeid, channelnum=channelnum, channelname=channelname, })
		tinsert(Settings.delaychannelupdate[action], { id=chatframeid, channelname=channelname, })
		return true
	end
end

-- ... = GetChatWindowChannels(chatframeId)
function NS:ListChannelNames(...)
	local List = {}
	for i=1, select("#", ...) do
		local Info = select(i, ...)
		if type(Info) ~= "number" then
			tinsert(List, Info)
		end
	end
	return #List > 0 and List or nil
end

function NS:RetstoreChatFrameSettings(restorefrom)
	for k, v in pairs(restorefrom) do
		local f = _G["ChatFrame"..k]
		if v.Messages then
			for i=1, #v.Messages do
				ChatFrame_AddMessageGroup(f, v.Messages[i])
			end
		end
		if v.Channels then
			for i=1, #v.Channels do
--				if not NS:ChannelUpdateDelayed(id, "add", Channels.Add[id][i].num, Channels.Add[id][i].name, GetChannelList()) then
				if not NS:ChannelUpdateDelayed(k, "add", v.Channels[i], GetChannelList()) then
					ChatFrame_AddChannel(f, v.Channels[i])
				end
			end
		end
		SetChatWindowSavedDimensions(k, v.Width, v. Height)
		if v.Point then
			SetChatWindowSavedPosition(k, v.Point, v.xOffset, v.yOffset)
		end
		NS:SetChatWindowName(k, v.Info[1])
		SetChatWindowSize(k, v.Info[2])
		SetChatWindowColor(k, v.Info[3], v.Info[4], v.Info[5])
		SetChatWindowAlpha(k, v.Info[6])
		SetChatWindowShown(k, v.Info[7])
		SetChatWindowLocked(k, v.Info[8])
		SetChatWindowDocked(k, v.Info[9])
		SetChatWindowUninteractable(k, v.Info[10])
	end
end

function NS:SaveChatFrameSettings(id, satevo)
	satevo[id] = {}
	satevo[id].Info = { GetChatWindowInfo(id) }
	local width, height = GetChatWindowSavedDimensions(id)
	satevo[id].Width = width
	satevo[id].Height = height
	local point, xOffset, yOffset = GetChatWindowSavedPosition(id)
	satevo[id].Point = point
	satevo[id].xOffset = xOffset
	satevo[id].yOffset = yOffset
	local Chat = { GetChatWindowMessages(id) }
	if #Chat > 0 then
		satevo[id].Messages = Chat
	end
	satevo[id].Channels = NS:ListChannelNames(GetChatWindowChannels(id))
end

